package prefuse.render;

import java.awt.Graphics2D;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;

import prefuse.Display;
import prefuse.util.ColorLib;
import prefuse.visual.EdgeItem;
import prefuse.visual.NodeItem;
import prefuse.visual.VisualItem;
import cytoPrefuse.CytoPrefuse;
import cytoPrefuse.EdgeBundler;

//Just like ShapeRenderer except we'll trigger mouse over events to an area twice as
//big as our object

//Also draws the starts of lines that join to this object over the object border in
// the color of this node.
public class EasyHoverShapeRenderer extends ExtendedShapeRenderer {
	 

	
	protected Line2D       m_line  = new Line2D.Float();	
	
	  private Display d;
	  private double scaleRange;
	  private double minScale;
	  public EasyHoverShapeRenderer( Display d) {
		  this.d = d;
		 
		  Rectangle2D bounds = d.getVisualization().getBounds(CytoPrefuse.NODES);
		  double w = d.getWidth();
		  double h = d.getHeight();
          double cx = bounds.getCenterX();
          double cy =  bounds.getCenterY();
          
          // compute half-widths of final bounding box around
          // the desired center point
          double wb = Math.max(cx-bounds.getMinX(),
                               bounds.getMaxX()-cx);
          double hb = Math.max(cy-bounds.getMinY(),
                               bounds.getMaxY()-cy);
          
          // compute scale factor
          //  - figure out if z or y dimension takes priority
          //  - then balance against the current scale factor
          minScale = Math.min(w/(2*wb),h/(2*hb));
          scaleRange = 1.5 - minScale;
	  }
	  public void setBounds(VisualItem item) {
		    super.setBounds(item);
		    Rectangle2D bounds = (Rectangle2D) item.get(VisualItem.BOUNDS);
		    double extraFactor =  1- ( (d.getScale() - minScale) / scaleRange);
		    if (extraFactor < 0)
		    	extraFactor = 0;
		    if (extraFactor> 1)
		    	extraFactor =1;
		    
		    double extraWidth = bounds.getWidth() *extraFactor;
		    double extraHeight = bounds.getHeight() *extraFactor;
		    bounds.setRect( bounds.getX() - extraWidth,
		    				bounds.getY() - extraHeight,
		    				bounds.getWidth() + 2*(extraWidth),
		    				bounds.getHeight() + 2 *(extraHeight));
		    
	    }
	   public boolean locatePoint(Point2D p, VisualItem item) {
	        if ( item.getBounds().contains(p) ) {
	        	return true;
	        }
	        return false;
	   }
	
	   public void render(Graphics2D g, VisualItem item) {		  	 
		   super.render(g,item);
		
		   List sortedEdges = new ArrayList();
		   RenderSorter renderSorter = new RenderSorter();
		   //Now draw coloured join lines over the edge of the shape
		   if (item instanceof NodeItem) {
			  NodeItem node = (NodeItem)item;
		
			  sortedEdges.clear();
			  Iterator edgesIt = node.edges();
		      while (edgesIt.hasNext()) {
		    	  sortedEdges.add(edgesIt.next());
		      }
		      Collections.sort(sortedEdges, renderSorter);
		     
		      edgesIt = sortedEdges.iterator();
		      while (edgesIt.hasNext()) {
		    	  EdgeItem edge = (EdgeItem) edgesIt.next();
			      if (edge.isVisible()){
			    	  g.setColor(ColorLib.getColor(edge.getStrokeColor()));
					  CytoPrefuse cytoPrefuse = (CytoPrefuse)d;
			    	  g.setStroke( cytoPrefuse.colorManager.edgeStrokeAction.getStroke(edge));			     
			    	  List controlPoints = (List)edge.get(CytoPrefuse.SPLINE_BETA_ADJUSTED);
			    
				    	  if (controlPoints.size() >0){
				    	  int edgeLength = edge.getInt(CytoPrefuse.EDGE_LENGTH);
				    	  float distanceToNodeEdge = (float) (node.getSize()*14.0f); 
				    	  float splinePercent = distanceToNodeEdge/ (float)edgeLength;
				    	  
				    	  float splineSegmentPercent = 1/ (float)EdgeBundler.SPLINE_SEGEMENT_COUNT;;
			
				    	  //round up
				    	  int segmentCount = (int) (splinePercent / splineSegmentPercent) +1;
				    	  
				    	  //If it's a self edge, draw it all
				    	  boolean selfEdge = false;
				    	  Point2D p1 = (Point2D)controlPoints.get(0);
				    	  Point2D pLast = (Point2D)controlPoints.get(controlPoints.size() - 1);
				    	  if (Math.abs(p1.getX() - pLast.getX()) < 0.001 && 
				    		Math.abs(p1.getY() - pLast.getY()) < 0.001){
				    		  segmentCount = EdgeBundler.SPLINE_SEGEMENT_COUNT;
				    		  selfEdge = true;
				    	  }
				    		
				    	  
				    	  //Either draw the first or last segment of the spline, 
				    	  //depending on which end of the edge this node is
				    	  int segmentPos;
				    	  int direction;
				    	  Point2D firstPoint = (Point2D)controlPoints.get(0);
				    	  double firstX;
				    	  double firstY;
				    	
				    	  if (Math.abs(firstPoint.getX() - node.getX()) < 0.1 &&
				    			  Math.abs(firstPoint.getY()-node.getY()) < 0.1) {
				    		  segmentPos = EdgeBundler.SPLINE_SEGMENT_START + EdgeBundler.SPLINE_SEGMENT_SIZE;
				    		  firstX = node.getX();
					    	  firstY = node.getY();			    	
				    		  direction = 1;
				    	  }else {
				    		  segmentPos = EdgeBundler.SPLINE_SEGMENT_END - EdgeBundler.SPLINE_SEGMENT_SIZE;
				    		  Point2D endPoint = EdgeBundler.getNextSplinePoint(controlPoints,EdgeBundler.SPLINE_SEGMENT_END);
				    		  firstX = endPoint.getX();
				    		  firstY = endPoint.getY();
					    			
				    		  direction = -1;
				    	  }
				    	  double prevX = firstX;
				    	  double prevY = firstY;
				    	  
				    	  for (int i=0; i < segmentCount; i++) {
				    		  
				    		  Point2D segment = EdgeBundler.getNextSplinePoint(controlPoints,segmentPos);
				    		 
				    		  //Ugh.  Sometimes even a single segment of spline overlaps the
				    		  //next node.  Check here, and adjust the spline point if that
				    		  //is the case
				    		  double deltax = (segment.getX() - firstX);
				    		  double deltay = (segment.getY() - firstY);
				    		  double length =  Math.sqrt (deltax*deltax + deltay*deltay);
				    		  if (length > distanceToNodeEdge && !selfEdge ){
				    			  float newSegmentLength = 1.0f - (distanceToNodeEdge / (float)(length));
				    			  segment.setLocation(segment.getX() - deltax*newSegmentLength,
				    					  			  segment.getY() - deltay*newSegmentLength);
				    			  
				    		  }
				    		  
				    		  
				    		  m_line.setLine(prevX,
					        		prevY,
					        		segment.getX(),
					        		segment.getY());
					        
				    		  prevX = segment.getX();
				    		  prevY = segment.getY();
				    		  
				    		  g.draw(m_line);
				    		  segmentPos += EdgeBundler.SPLINE_SEGMENT_SIZE * direction;
				    	  };
					  }
			      }
			   }			  
		   }
	   }
	   
	   private class RenderSorter implements Comparator {

			public int compare(Object arg0, Object arg1) {
				EdgeItem e1 = (EdgeItem)arg0;
				EdgeItem e2 = (EdgeItem)arg1;
				if (e1.getBoolean(CytoPrefuse.ON_SHORTEST_PATH))
					return 1;
				if (e2.getBoolean(CytoPrefuse.ON_SHORTEST_PATH))
					return -1;
				
				if (e1.isInGroup(CytoPrefuse.SINGLE_HIGHLIGHT))
					return 1;
				if (e2.isInGroup(CytoPrefuse.SINGLE_HIGHLIGHT))
					return -1;
							
					
					if (e1.getBoolean(VisualItem.HOVER)) 
							return 1;
					if (e2.getBoolean(VisualItem.HOVER)) 
							return -1;
							
					if (e1.getBoolean(VisualItem.HIGHLIGHT))
							return 1;
					if (e2.getBoolean(VisualItem.HIGHLIGHT))
							return -1;								
					if (e1.isInGroup(CytoPrefuse.SELECTED))
							return 1;	
					if (e2.isInGroup(CytoPrefuse.SELECTED))
						return -1;	
				
					if (e1.getInt(CytoPrefuse.EDGE_LENGTH) < e2.getInt(CytoPrefuse.EDGE_LENGTH)){
						return 1;
					}else				
						return -1;		
			}

		}
}
